""" The loopy specific generators """
from __future__ import absolute_import

from dune.perftool.generation import (generator_factory,
                                      no_caching,
                                      preamble,
                                      )

import loopy
import numpy

iname = generator_factory(item_tags=("iname",))
pymbolic_expr = generator_factory(item_tags=("kernel", "pymbolic"))
function_mangler = generator_factory(item_tags=("mangler",))


@generator_factory(item_tags=("argument", "globalarg"),
                   cache_key_generator=lambda n, **kw: n)
def globalarg(name, shape=loopy.auto, **kw):
    if isinstance(shape, str):
        shape = (shape,)
    dtype = kw.pop("dtype", numpy.float64)
    return loopy.GlobalArg(name, dtype=dtype, shape=shape, **kw)


@generator_factory(item_tags=("argument", "constantarg"),
                   cache_key_generator=lambda n, **kw: n)
def constantarg(name, shape=loopy.auto, **kw):
    if isinstance(shape, str):
        shape = (shape,)
    dtype = kw.pop("dtype", numpy.float64)
    return loopy.GlobalArg(name, dtype=dtype, shape=shape, **kw)


@generator_factory(item_tags=("argument", "valuearg"),
                   cache_key_generator=lambda n, **kw: n)
def valuearg(name, **kw):
    return loopy.ValueArg(name, **kw)


@generator_factory(item_tags=("domain",))
def domain(iname, shape):
    if isinstance(shape, str):
        valuearg(shape)
    return "{{ [{0}] : 0<={0}<{1} }}".format(iname, shape)


def _temporary_type(shape_impl, shape, first=True):
    if len(shape_impl) == 0:
        return 'double'
    if shape_impl[0] == 'arr':
        if not first or len(set(shape_impl)) != 1:
            raise ValueError("We do not allow mixing of C++ containers and plain C arrays, for reasons of mental sanity")
        return 'double'
    if shape_impl[0] == 'vec':
        return "std::vector<{}>".format(_temporary_type(shape_impl[1:], shape[1:], first=False))
    if shape_impl[0] == 'fv':
        return "Dune::FieldVector<{}, {}>".format(_temporary_type(shape_impl[1:], shape[1:], first=False), shape[0])
    if shape_impl[0] == 'fm':
        # For now, no field matrices of weird stuff...
        assert len(shape) == 2
        return "Dune::FieldMatrix<double, {}, {}>".format(shape[0], shape[1])


@preamble
def default_declaration(name, shape=(), shape_impl=()):
    # Determine the C++ type to use for this temporary.
    t = _temporary_type(shape_impl, shape)
    if len(shape_impl) == 0:
        # This is a scalar, just return it!
        return '{} {}(0.0);'.format(t, name)
    if shape_impl[0] == 'arr':
        return '{} {}{};'.format(t, name, ''.join('[{}]'.format(s) for s in shape))
    if shape_impl[0] == 'vec':
        return '{} {}({});'.format(t, name, shape[0])
    if shape_impl[0] == 'fv':
        return '{} {}(0.0);'.format(t, name)
    if shape_impl[0] == 'fm':
        assert len(shape) == 2
        return '{} {}(0.0);'.format(t, name)


def declaration_with_base_storage(base_storage):
    @preamble
    def base_storage_decl(name, shape=(), shape_impl=()):
        return 'double* {} = {};'.format(name, base_storage)

    return base_storage_decl


class _TemporaryCounter:
    counter = 0


def get_temporary_name():
    name = 'expr_{}'.format(str(_TemporaryCounter.counter).zfill(4))
    _TemporaryCounter.counter = _TemporaryCounter.counter + 1
    return name


class UnmanagedTemporaryVariable(loopy.TemporaryVariable):
    """
    A TemporaryVariable class that does not have knowledge of data layout.
    Is used when the data layout is under control of the underlying C++ type.
    Indexed access will result in multiple operator[]s. Only known to
    the DuneTarget.
    """
    pass


@generator_factory(item_tags=("temporary",), cache_key_generator=no_caching)
def temporary_variable(name, managed=False, **kwargs):
    if 'dtype' not in kwargs:
        kwargs['dtype'] = numpy.float64

    shape = kwargs.get('shape', ())
    shape_impl = kwargs.pop('shape_impl', ('arr',) * len(shape))
    decl_method = kwargs.pop('decl_method', default_declaration)

    # Check if a specified base_storage has already been initialized
    if kwargs.get('base_storage', None):
        assert(kwargs['base_storage'] in [tv.name for tv in temporary_variable._memoize_cache.values()])
        assert(decl_method is None)
        decl_method = declaration_with_base_storage(kwargs['base_storage'])

    if decl_method is not None:
        decl_method(name, shape, shape_impl)

    if managed:
        return loopy.TemporaryVariable(name, **kwargs)
    else:
        return UnmanagedTemporaryVariable(name, **kwargs)


# Now define generators for instructions. To ease dependency handling of instructions
# these generators are a bit more involved... We apply the following procedure:
# There is one generator that returns the unique id and forwards to a generator that
# actually adds the instruction. Hashing is done based on the code snippet.


@generator_factory(item_tags=("instruction", "cinstruction"),
                   cache_key_generator=lambda *a, **kw: kw['code'],
                   )
def c_instruction_impl(**kw):
    kw.setdefault('assignees', [])
    from pymbolic.primitives import Variable
    kw['assignees'] = frozenset(Variable(i) for i in kw['assignees'])
    inames = kw.pop('inames', kw.get('forced_iname_deps', []))

    return loopy.CInstruction(inames, **kw)


@generator_factory(item_tags=("instruction", "exprinstruction"),
                   cache_key_generator=lambda *a, **kw: kw['expression'],
                   )
def expr_instruction_impl(**kw):
    if 'assignees' in kw:
        from pymbolic.primitives import Variable
        kw['assignees'] = frozenset(Variable(i) for i in kw['assignees'])
    return loopy.ExpressionInstruction(**kw)


@generator_factory(item_tags=("instruction", "callinstruction"),
                   cache_key_generator=lambda *a, **kw: kw['expression'],
                   )
def call_instruction_impl(**kw):
    return loopy.CallInstruction(**kw)


class _IDCounter:
    count = 0


def _insn_cache_key(code=None, expression=None, **kwargs):
    if code is not None:
        return code
    if expression is not None:
        return expression
    raise ValueError("Please specify either code or expression for instruction!")


@generator_factory(item_tags=("insn_id",), cache_key_generator=_insn_cache_key)
def instruction(code=None, expression=None, **kwargs):
    assert (code is not None) or (expression is not None)
    assert not ((code is not None) and (expression is not None))
    assert 'id' not in kwargs

    # Get an ID for this instruction
    id = 'insn' + str(_IDCounter.count).zfill(4)
    _IDCounter.count = _IDCounter.count + 1

    # Now create the actual instruction
    if code:
        c_instruction_impl(id=id, code=code, **kwargs)
    if expression:
        if 'assignees' in kwargs and len(kwargs['assignees']) == 0:
            call_instruction_impl(id=id, expression=expression, **kwargs)
        else:
            expr_instruction_impl(id=id, expression=expression, **kwargs)

    # return the ID, as it is the only useful information to the user
    return id
